"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import io
import json
import warnings
from copy import copy, deepcopy
from pathlib import Path
from threading import Lock
from typing import Callable

import math
import pandas as pd
import zarr
from cachetools import LRUCache
from compress_pickle import dump, load

# TODO: can we implement our own hash function for deskriptors to reduce dependency on dask?
from dask.base import tokenize

from .core import task
from .deskribe import Range
from .utils import (
    NodeFailedException,
    dict_update,
    exclusive_indexing,
    indexers_to_slices,
)


def to_pickle_with_store(store, filename, object, compression="gzip"):
    bytes_buffer = io.BytesIO()
    dump(object, bytes_buffer, compression=compression)
    store[filename] = bytes_buffer.getvalue()


def read_pickle_with_store(store, filename, compression="gzip"):
    bytes_buffer = io.BytesIO(store[str(filename)])
    return load(bytes_buffer, compression=compression)


def string_timestamp(o):
    if hasattr(o, "isoformat"):
        return o.isoformat()
    else:
        return str(o)


@task()
class HashPersister:
    def __init__(
        self,
        store=None,
        selected_keys=None,
        compression=None,
        force_update=False,
        use_memorycache=True,
    ):
        super().__init__(force_update=force_update, use_memorycache=use_memorycache)
        if isinstance(store, str) or isinstance(store, Path):
            store = zarr.DirectoryStore(store)
        self.store = store
        self.compression = compression
        self._mutex = Lock()
        # self.what_is_being_written = manager.dict()

        if selected_keys is None:
            # use all keys as hash
            pass

        self.cache = LRUCache(10)

    def get_hash(self, request):
        # ignore all configs that are meant for hashpersister
        r = {k: v for k, v in request.items() if k != "self"}
        s = json.dumps(r, sort_keys=True, skipkeys=True, default=string_timestamp)
        request_hash = tokenize(s)

        return request_hash

    def is_valid(self, request):
        """Checks if persisted object for `request`
        exists and is valid (i.e. is not of type NodeFailedException).

        Args:
            request (dict): The request that should be checked

        Returns:
            boolean or None: Returns false if the persisted item is of type NodeFailedException
                             Returns None if the request has not been persisted yet.
        """
        request_hash = self.get_hash(request)

        if "fail/" + request_hash in self.store:
            return False

        if request_hash in self.store:
            return True

        return None

    def configure(self, request=None):
        request_hash = self.get_hash(request)

        # compute action defaults to passthrough
        request["self"]["action"] = "passthrough"

        if request["self"].get("bypass", False):
            # set to passthrough -> nothing will happen
            return request

        # propagate the request_hash to the compute function
        request["self"]["request_hash"] = request_hash

        # reload and rewrite the chunk if requested
        if request["self"].get("force_update", False):
            request["self"]["action"] = "store"
            return request

        with self._mutex:
            if (
                request["self"].get("use_memorycache", True)
                and request_hash in self.cache
            ):
                request["remove_dependencies"] = True
                # set the compute action to load
                request["self"]["action"] = "load_from_cache"
                return request

            # while holding the mutex, we need to check if the file exists
            if request_hash in self.store:
                # remove previous node since we are going to load from disk
                request["remove_dependencies"] = True

                # set the compute action to load
                request["self"]["action"] = "load"
                return request

            if "fail/" + request_hash in self.store:
                # remove previous node since we are going to load the fail info from disk
                request["remove_dependencies"] = True
                request["self"]["request_hash"] = "fail/" + request_hash

                # set the compute action to load
                request["self"]["action"] = "load"
                return request

            # # check if the file will be written to already
            # if request_hash in self.what_is_being_written:
            #     # yes? that's fine another process handled the same request
            #     # we must tell the system to load the data regularly
            #     # otherwise this node might not get data if the write wasn't
            #     # finished before this node's compute call is processed
            #     # let the system take care of optimizing potential double computations
            #     return request

            # # register that we are going to write to a file
            # self.what_is_being_written[request_hash] = True
            request["self"]["action"] = "store"

        return request

    def compute(self, data=None, **kwargs):
        request = kwargs
        if request["action"] == "load_from_cache":
            with self._mutex:
                cached = self.cache[request["request_hash"]]
            return cached
        if request["action"] == "load":
            data = read_pickle_with_store(
                self.store,
                request["request_hash"],
                compression=self.compression,
            )

            if isinstance(data, str):
                raise RuntimeError(f"something wrong read {data}")

            with self._mutex:
                self.cache[request["request_hash"]] = data

            return data
        elif request["action"] == "store":
            try:
                self.cache[request["request_hash"]] = data

                # write to file
                # TODO: write to tmp file and move in place
                if isinstance(data, NodeFailedException):
                    to_pickle_with_store(
                        self.store,
                        "fail/" + request["request_hash"],
                        data,
                        compression=self.compression,
                    )
                else:
                    if isinstance(data, str):
                        raise RuntimeError(f"something wrong {data}")
                    to_pickle_with_store(
                        self.store,
                        request["request_hash"],
                        data,
                        compression=self.compression,
                    )

            except Exception as e:
                print("Error during hashpersister", repr(e))
            # finally:
            #     with self._mutex:
            #         # de-register this hash
            #         del self.what_is_being_written[request["request_hash"]]
            with self._mutex:
                self.cache[request["request_hash"]] = data

            return data
        elif request["action"] == "passthrough":
            return data
        else:
            raise NodeFailedException("A bug in HashPersister. Please report.")


def to_datetime(x, **kwargs):
    # overwrites default
    utc = kwargs.pop("utc", True)
    if not utc:
        warnings.warn(
            "to_datetime overwrites your keyword utc argument and enforces `utc=True`"
        )
    return pd.to_datetime(x, utc=True, **kwargs).tz_localize(None)


def is_datetime(x):
    return pd.api.types.is_datetime64_any_dtype(x)


def to_datetime_conditional(x, condition=True, **kwargs):
    # converts x to datetime if condition is true or the object in condition is datetime or timedelta
    if not isinstance(condition, bool):
        condition = is_datetime(condition) or isinstance(condition, pd.Timedelta)

    if condition:
        return to_datetime(x, **kwargs)
    return x


def get_segments(
    dataset_scope,
    segment_slice,
    segment_stride=None,
    reference=None,
    mode="overlap",
    minimal_number_of_segments=0,
    timestamps_as_strings=False,
    utc_no_tz=True,
):
    # modified from and thanks to xbatcher: https://github.com/rabernat/xbatcher/
    if isinstance(mode, str):
        mode = {dim: mode for dim in segment_slice}

    if segment_stride is None:
        segment_stride = {}

    if reference is None:
        reference = {}

    dim_slices = []
    dims = []
    for dim in segment_slice:
        if dim not in dataset_scope:
            continue
        dims += [dim]

        _segment_slice = segment_slice[dim]
        _segment_stride = segment_stride.get(dim, _segment_slice)

        dataset_scope_dim = dataset_scope[dim]
        if isinstance(dataset_scope_dim, list):
            segment_start = 0
            segment_end = len(dataset_scope_dim)

            if _segment_slice == "full":
                dim_slices += [[dataset_scope_dim]]
                continue

        elif isinstance(dataset_scope[dim], (Range, dict)):
            if isinstance(dataset_scope[dim], Range):
                dataset_scope_dim = dict(dataset_scope[dim])

            if _segment_slice == "full":
                dim_slices += [[dataset_scope_dim]]
                continue

            segment_start = to_datetime_conditional(
                dataset_scope_dim["start"], segment_slice[dim]
            )
            segment_end = to_datetime_conditional(
                dataset_scope_dim["end"], segment_slice[dim]
            )

            if mode[dim] == "overlap":
                # TODO: add options for closed and open intervals
                # first get the lowest that window that still overlaps with our segment
                segment_start = (
                    segment_start
                    - math.floor(_segment_slice / _segment_stride) * _segment_stride
                )
                # then align to the grid if necessary
                if dim in reference:
                    ref_dim = to_datetime_conditional(
                        reference[dim], segment_slice[dim]
                    )
                    segment_start = (
                        math.ceil((segment_start - ref_dim) / _segment_stride)
                        * _segment_stride
                        + ref_dim
                    )
                segment_end = segment_end + _segment_slice
            elif mode[dim] == "fit":
                if dim in reference:
                    ref_dim = to_datetime_conditional(
                        reference[dim], segment_slice[dim]
                    )
                    segment_start = (
                        math.floor((segment_start - ref_dim) / _segment_stride)
                        * _segment_stride
                        + ref_dim
                    )
                else:
                    raise RuntimeError(
                        f"mode `fit` requires that dimension {dim} is in reference {reference}"
                    )
            else:
                RuntimeError(f"Unknown mode {mode[dim]}. It must be `fit` or `overlap`")

        if isinstance(
            segment_slice[dim], pd.Timedelta
        ):  # or isinstance(segment_slice[dim], dt.timedelta):
            # TODO: change when xarray #3291 is fixed
            iterator = pd.date_range(segment_start, segment_end, freq=_segment_stride)
            segment_end = pd.to_datetime(segment_end)
        else:
            iterator = range(segment_start, segment_end, _segment_stride)

        slices = []
        for start in iterator:
            end = start + _segment_slice
            if end <= segment_end or (
                len(slices) < minimal_number_of_segments
                and not isinstance(dataset_scope_dim, list)
            ):
                if is_datetime(start):
                    if utc_no_tz:
                        start = pd.to_datetime(start, utc=True).tz_localize(None)
                    if timestamps_as_strings:
                        start = start.isoformat()
                if is_datetime(end):
                    if utc_no_tz:
                        end = pd.to_datetime(end, utc=True).tz_localize(None)
                    if timestamps_as_strings:
                        end = end.isoformat()

                if isinstance(dataset_scope_dim, list):
                    slices.append(dataset_scope_dim[start:end])
                else:
                    slices.append({"start": start, "end": end})

        dim_slices.append(slices)

    import itertools

    all_slices = []
    for slices in itertools.product(*dim_slices):
        selector = {key: slice for key, slice in zip(dims, slices)}
        all_slices.append(selector)

    return all_slices


try:
    import xarray as xr
except ImportError:
    warnings.warn("Install xarray to use the default merge function of ChunkPersister")


def merge_xarray(data, request):
    data = [d for d in data if d is not None]
    for i in range(len(data)):
        if hasattr(data[i], "name") and (not data[i].name or data[i].name is None):
            data[i].name = "data"
    merged_dataset = xr.merge(data)
    # if hasattr(data[0], "name"):
    #     merged_dataset = merged_dataset[data[0].name]
    indexers = {}
    for coord in merged_dataset.coords:
        if coord in request:
            indexers[coord] = request[coord]
    slices = indexers_to_slices(indexers)
    section = merged_dataset.sel(slices)
    section = exclusive_indexing(section, indexers)
    return section


@task()
class ChunkPersister:
    def __init__(
        self,
        store,
        dim: str = "time",
        # classification_scope:dict | Callable[...,dict]=None,
        segment_slice: dict | Callable[..., dict] = None,
        # segment_stride:dict|Callable[...,dict]=None,
        dataset_scope: dict | Callable[..., dict] = None,
        mode: str = "overlap",
        reference: dict = None,
        force_update=False,
        merge_function=None,
    ):
        """Chunks every incoming dekriptor into subchunks if deskriptor is larger than segment_slice
         or extends the deskriptor to the respective chunksize if deskriptor is smaller than segment_slice

        Args:
            store (_type_): _description_
            dim (str, optional): _description_. Defaults to "time".
            segment_slice (dict | Callable[..., dict], optional): A dictionary containing an entry for each dimension that should be chunked. Each entry is the respective chunk size given in the units of the expected dimension of the deskriptor. For example, for a time dimension you can use pd.Timedelta. Defaults to None.
            dataset_scope (dict | Callable[...,dict], optional): The extend of the chunking. If None, the incoming deskriptor will be used as the scope. If only selected dimensions are given as dataset_scope, the scope for the other dimensions will be choosen from the incoming deskriptor. Defaults to None.
            mode (str, optional): _description_. Defaults to "overlap".
            reference (dict, optional): _description_. Defaults to None.
            force_update (bool, optional): _description_. Defaults to False.
            merge_function (_type_, optional): _description_. Defaults to None.
        """
        # if callable(classification_scope):
        #     self.classification_scope = classification_scope
        #     classification_scope = None
        # else:
        #     self.classification_scope = None

        if callable(segment_slice):
            self.segment_slice = segment_slice
            segment_slice = None
        else:
            self.segment_slice = None

        # if callable(segment_stride):
        #     self.segment_stride = segment_stride
        #     segment_stride = None
        # else:
        #     self.segment_stride = None

        self.merge = merge_function
        if self.merge is None:
            self.merge = merge_xarray

        super().__init__(
            dim=dim,
            # classification_scope=classification_scope,
            segment_slice=segment_slice,
            # segment_stride=segment_stride,
            dataset_scope=dataset_scope,
            mode=mode,
            reference=reference,
            force_update=force_update,
        )

        if isinstance(store, str) or isinstance(store, Path):
            store = zarr.DirectoryStore(store)
        self.store = store

    def __dask_tokenize__(self):
        return (ChunkPersister,)

    def configure(self, request=None):
        rs = request["self"]

        def get_value(attr_name):
            # decide if we use the attribute provided in the request or
            # from a callback provided at initialization

            if rs.get(attr_name, None) is None:
                # there is no attribute in the request, check for callback
                callback = getattr(self, attr_name)
                if callback is not None and callable(callback):
                    value = callback(request)
                else:
                    RuntimeError("No valid classification_scope provided")
            else:
                value = rs[attr_name]
            return value

        dataset_scope = copy(rs)
        if rs.get("dataset_scope", None) is not None:
            dataset_scope.update(rs["dataset_scope"])
        segment_slice = get_value("segment_slice")
        # segment_stride = get_value("segment_stride")
        segments = get_segments(
            dataset_scope,
            segment_slice,
            # segment_stride,
            reference=rs["reference"],
            mode=rs["mode"],
            timestamps_as_strings=True,
            minimal_number_of_segments=1,
        )
        cloned_requests = []
        cloned_hashpersisters = []
        for segment in segments:
            segment_request = deepcopy(request)
            if "self" in segment_request:
                del segment_request["self"]
            dict_update(segment_request, segment)
            cloned_requests += [segment_request]
            cloned_hashpersister = HashPersister(
                store=self.store,
            )
            cloned_hashpersister.dask_key_name = self.dask_key_name + "_hashpersister"
            dict_update(
                segment_request,
                {
                    "config": {
                        "keys": {
                            self.dask_key_name + "_hashpersister": {
                                "force_update": rs.get("force_update", False)
                            }
                        }
                    }
                },
            )
            cloned_hashpersisters += [cloned_hashpersister.compute]

        # Insert predecessor
        # new_request = {}
        request["clone_dependencies"] = cloned_requests
        request["insert_predecessor"] = cloned_hashpersisters

        return request

    def compute(self, *data, **request):
        def unpack_list(inputlist):
            new_list = []
            for item in inputlist:
                if isinstance(item, list):
                    new_list += unpack_list(item)
                else:
                    new_list += [item]
            return new_list

        data = unpack_list(data)
        success = [d for d in data if not isinstance(d, NodeFailedException)]

        if not success:
            failed = [str(d) for d in data if isinstance(d, NodeFailedException)]
            raise RuntimeError(f"Failed to load data. Reason: {failed}")
        section = self.merge(success, request)
        return section



"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import importlib
import inspect

from .core import it, task
from .graph import Node, compute
from .persist import ChunkPersister, HashPersister


def deklare_flow(flow):
    flow_graph = it(flow)

    def deklare_flow_function(self, deskriptor):
        return compute(flow_graph, deskriptor)

    def query(self, deskriptor):
        return compute(flow_graph, deskriptor)

    if hasattr(flow, "__self__") and flow.__self__ is not None:
        flow.query = query
    else:
        flow.query = lambda deskriptor: compute(flow_graph, deskriptor)

    return flow


def is_module_function_or_class(module):
    def predicate(member):
        # TODO: should we also check for member.__name__ in module.__name__?
        return inspect.isfunction(member) or inspect.isclass(member)

    return predicate


def deklare_module(
    module, flows=None, names=None, external_tasks=None, ignore=None, no_wrap=False
):
    names = names or {}
    external_tasks = external_tasks or []
    ignore = ignore or []

    flows = flows or set()
    if isinstance(flows, str):
        flows = {flows}

    flows = set(flows)

    if isinstance(module, str):
        # Dynamically import the module
        module = importlib.import_module(module)

    # Create a dictionary to store original functions that should not be decorated
    flow_instances = {}

    # dependencies
    dependency_modules = {}

    # Iterate over all functions defined in the module
    for name, func_or_cls in inspect.getmembers(
        module, predicate=is_module_function_or_class(module)
    ):
        if name in flows:
            # if member direct member of the module, we add it to the return flows
            # if not, it needs to be loaded appropriately from it's original module
            if func_or_cls.__module__ == module.__name__:
                flow_instances[name] = func_or_cls
            else:
                # select to load it as a deklare dependency, i.e. load all required members
                # as deklare tasks
                dependency_modules[func_or_cls.__module__] = dependency_modules.get(
                    func_or_cls.__module__, {}
                ) | {name: func_or_cls}
        elif name not in ignore and (
            func_or_cls.__module__ == module.__name__ or name in external_tasks
        ):
            # Decorate the function and add it back to the module's namespace
            key_names = names.get(name, None)
            setattr(module, name, task(name=key_names)(func_or_cls))

    for dependency_module, dependency_flows in dependency_modules.items():
        dependency_flows_names = list(dependency_flows.keys())
        module_and_flows = deklare_module(
            dependency_module, flows=dependency_flows_names, no_wrap=True
        )
        for i, flow in enumerate(module_and_flows[1:]):
            setattr(module, flow.__name__, flow)
            # # TODO: would the above fail if we use `from module import flow as f`?
            # # maybe better to do? but then we need to make sure that it's always the same order, because it's probably not!
            # setattr(module, dependency_flows[i], flow)

    for name, func_or_cls in flow_instances.items():
        if no_wrap:
            flow_instances[name] = func_or_cls
        else:
            # if inspect.isclass(func_or_cls):
            #     func_or_cls = func_or_cls()
            flow_instances[name] = deklare_flow(func_or_cls)

    return tuple([module] + list(flow_instances.values()))



"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import dask
from dask.delayed import Delayed

from .graph import base_name


def indexers_to_slices(indexers):
    new_indexers = {}
    for key in indexers:
        if isinstance(indexers[key], dict):
            ni = {"start": None, "end": None, "step": None}
            ni.update(indexers[key])
            new_indexers[key] = slice(ni["start"], ni["end"], ni["step"])
        else:
            new_indexers[key] = indexers[key]

    return new_indexers


def exclusive_indexing(x, indexers):
    # Fake `exlusive indexing`
    drop_indexers = {k: indexers[k]["end"] for k in indexers if "end" in indexers[k]}
    try:
        x = x.drop_sel(drop_indexers, errors="ignore")
    except Exception:
        pass

    return x


class NodeFailedException(Exception):
    def __init__(self, exception=None):
        """The default exception when a node's compute function fails and failsafe mode
        is enable, i.e. the global setting `fail_mode` is not set to `fail`.
        this exception is caught by the foreal processing system and depending on the
        global variable `fail_mode`, leads to process interruption or continuation.

        Args:
            exception (any, optional): The reason why it failed, e.g. another exception.
                Defaults to None.
        """
        # if get_setting("fail_mode") == "warning" or get_setting("fail_mode") == "warn":
        #     print(exception)
        self.exception = exception

    def __str__(self):
        return str(self.exception)


def dict_update(base, update):
    if not isinstance(base, dict) or not isinstance(update, dict):
        raise TypeError(
            f"dict_update requires two dicts as input. But we received {type(base)} and {type(update)}"
        )

    for key in update:
        if isinstance(base.get(key), dict) and isinstance(update[key], dict):
            base[key] = dict_update(base[key], update[key])
        else:
            base[key] = update[key]

    return base


def extract_subgraphs(taskgraph, keys, match_base_name=False):
    if not isinstance(taskgraph, list):
        taskgraph = [taskgraph]

    extracted_graph, ck = dask.base._extract_graph_and_keys(taskgraph)
    if match_base_name:
        configured_graph_keys = list(extracted_graph.keys())
        new_keys = []
        for k in configured_graph_keys:
            for sk in keys:
                if base_name(sk) == base_name(k):
                    new_keys += [k]
        keys = new_keys
    return Delayed(keys, extracted_graph)



"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import datetime
from typing import Annotated, Any, Dict, Generic, TypeVar, get_args

import numpy as np
import pandas as pd
from pydantic import BaseModel, model_validator
from pydantic.functional_validators import AfterValidator


def to_datetime(v: Any) -> Any:
    return pd.to_datetime(v, utc=True).tz_localize(None)


T = TypeVar("T", str, datetime.datetime, datetime.date, int, float, type(np.datetime64))
DateTimeType = Annotated[T, AfterValidator(to_datetime)]


# class DatetimeRange(BaseModel):
#     start: DateTimeType
#     end: DateTimeType


T = TypeVar("T", int, float, DateTimeType)


class Range(BaseModel, Generic[T]):
    start: T
    end: T


DatetimeRange = Range[DateTimeType]


class Deskriptor(BaseModel, validate_assignment=True):
    @model_validator(mode="before")
    def dynamic_validator(cls, values: dict[str, Any]) -> dict[str, Any]:
        kwargs = {}
        for key, value in values.items():
            if key in cls.model_fields:
                field_info = cls.model_fields[key]
                if DatetimeRange in get_args(field_info.annotation) and isinstance(
                    value, dict
                ):
                    kwargs[key] = DatetimeRange(**value)
                elif Range in get_args(field_info.annotation) and isinstance(
                    value, dict
                ):
                    kwargs[key] = Range(**value)
                else:
                    kwargs[key] = value
        return kwargs

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        kwargs = {}
        for key, value in data.items():
            if key in cls.model_fields:
                field_info = cls.model_fields[key]
                if DatetimeRange in get_args(field_info.annotation) and isinstance(
                    value, dict
                ):
                    kwargs[key] = DatetimeRange(**value)
                elif Range in get_args(field_info.annotation) and isinstance(
                    value, dict
                ):
                    kwargs[key] = Range(**value)
                else:
                    kwargs[key] = value

        return cls(**kwargs)



"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import traceback
from copy import copy, deepcopy
from sqlite3 import InternalError

import dask
from dask.base import tokenize
from dask.core import flatten, get_dependencies, reverse_dict
from dask.delayed import Delayed
from dask.optimization import cull, fuse, inline
from dask.utils import (
    apply,
)

from .core import KEY_SEP, Node


def base_name(name):
    return name.split(KEY_SEP)[0]


FUNCTION = 1
DATA = 2
REQUEST = 3

# FUNCTION = 0
# DATA = 1
# REQUEST = 2


class App:
    __conf = {
        "fail_mode": "fail",
        "use_delayed": False,
    }
    __setters = ["fail_mode", "use_delayed"]

    @staticmethod
    def config(name):
        return App.__conf[name]

    @staticmethod
    def exists(name):
        return name in App.__conf

    @staticmethod
    def set(name, value):
        if name in App.__setters:
            App.__conf[name] = value
        else:
            raise NameError("Name not accepted in set() method")


def setting_exists(key):
    return App.exists(key)


def get_setting(key):
    return App.config(key)


def set_setting(key, value):
    return App.set(key, value)


def dict_update(base, update, convert_nestedfrozen=False):
    if not isinstance(base, dict) or not isinstance(update, dict):
        raise TypeError(
            f"dict_update requires two dicts as input. But we received {type(base)} and {type(update)}"
        )

    for key in update:
        if isinstance(base.get(key), dict) and isinstance(update[key], dict):
            if convert_nestedfrozen:
                base[key] = dict(base[key])
            base[key] = dict_update(
                base[key], update[key], convert_nestedfrozen=convert_nestedfrozen
            )
        else:
            base[key] = update[key]

    return base


class NodeFailedException(Exception):
    def __init__(self, exception=None):
        """The default exception when a node's compute function fails and failsafe mode
        is enable, i.e. the global setting `fail_mode` is not set to `fail`.
        this exception is caught by the deklare processing system and depending on the
        global variable `fail_mode`, leads to process interruption or continuation.

        Args:
            exception (any, optional): The reason why it failed, e.g. another exception.
                Defaults to None.
        """
        if get_setting("fail_mode") == "warning" or get_setting("fail_mode") == "warn":
            print(exception)
        self.exception = exception

    def __str__(self):
        return str(self.exception)


class FailSafeWrapper:
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        try:
            return self.func(*args, **kwargs)
        except Exception:
            trace = traceback.format_exc(2)
            return NodeFailedException(trace)


def update_key_in_config(request, old_key, new_key):
    if "config" in request:
        if "keys" in request["config"]:
            if old_key in request["config"]["keys"]:
                request["config"]["keys"][new_key] = request["config"]["keys"].pop(
                    old_key
                )


def generate_clone_key(current_node_name, to_clone_key, clone_id):
    return (
        base_name(to_clone_key)
        + KEY_SEP
        + tokenize([current_node_name, to_clone_key, "deklare_clone", clone_id])
    )


# @profile
def configuration(
    delayed,
    request,
    keys=None,
    default_merge=None,
    optimize_graph=True,
    dependants=None,
    clone_instead_merge=True,
):
    """Configures each node of the graph by propagating the request from outputs
    to inputs. Each node checks if it can fulfil the request and what it needs to fulfil
    the request. If a node requires additional configurations to fulfil the request it
    can set the 'requires_request' flag in the returned request and this function will
    add the return request as a a new input to the node's __call__().
    See also Node.configure()

    Args:
        delayed (dask.delayed or list): Delayed object or list of delayed objects
        request (dict or list): request (dict), list of requests
        keys (_type_, optional): _description_. Defaults to None.
        default_merge (_type_, optional): _description_. Defaults to None.
        optimize_graph (bool, optional): _description_. Defaults to True.
        dependants (_type_, optional): _description_. Defaults to None.

    Raises:
        RuntimeError: If graph cannot be configured

    Returns:
        dask.delayed: The configured graph
    """

    if not isinstance(delayed, list):
        collections = [delayed]
    else:
        collections = delayed

    # dsk = dask.base.collections_to_dsk(collections)
    dsk, dsk_keys = dask.base._extract_graph_and_keys(collections)
    if dependants is None:
        _, dependants = dask.core.get_deps(dsk)

    # dsk_dict = {k:dsk[k] for k in dsk.get_all_external_keys()}
    dsk_dict = {k: dsk[k] for k in dsk.keys()}

    if keys is None:
        keys = dsk_keys
    if not isinstance(keys, (list, set)):
        keys = [keys]

    work = list(set(flatten(keys)))
    # create a deepcopy, otherwise we might overwrite requests and falsify its usage outside of this function
    # request = deepcopy(request)
    if isinstance(request, list):
        # request = [NestedFrozenDict(r) for r in request if r]
        request = [r for r in request if r]
        if len(request) != len(work):
            raise RuntimeError(
                "When passing multiple request items "
                "The number of request items must be same "
                "as the number of keys"
            )

        # For each output node different request has been provided
        requests = {work[i]: [request[i]] for i in range(len(request))}
    else:
        # request = NestedFrozenDict(request)
        # Every output node receives the same request
        requests = {k: [request] for k in work}

    remove = {k: False for k in work}
    input_requests = {}
    # We will create a new graph with the configured nodes of the old graph
    # out_keys keeps track of the keys we have configured and
    # remember them for assembling the new graph
    out_keys = []

    # using dict here for performance because we are doing `if key in work` later
    # (and not using sets because for unknown reasons it doesn't work)
    work = {k: True for k in work}

    def normalize_node(k):
        # any node that does not have the following structure will get it
        # (apply, func, args, kwargs)
        if dsk_dict[k][0] is not apply:
            dsk_dict[k] = (apply, dsk_dict[k][0], list(dsk_dict[k][1:]), {})

    while work:
        # new_work = []
        new_work = dict()

        out_keys += work
        for k in work:
            # if k not in requests:
            #     # there wasn't any request stored use initial config
            #     requests[k] = [config]

            # check if we have collected all dependencies so far
            # we will come back to this node another time
            # TODO: make a better check for the case when dependants[k] is a set, also: why is it a set in the first place..?
            if (
                k in dependants
                # and len(dependants[k]) != len(requests[k])
                and not isinstance(dependants[k], set)
            ):
                continue

            if k not in requests:
                InternalError(f"Failed to find request for node {k}")

            # set configuration for this node k
            argument_is_node = None
            if isinstance(dsk_dict[k], tuple):
                # any node that does not have the following structure will get it
                # (apply, func, args, kwargs)
                normalize_node(k)

                # now every node will have the function in the second item of the tuple
                if hasattr(dsk_dict[k][FUNCTION], "__self__"):
                    if isinstance(dsk_dict[k][FUNCTION].__self__, Node):
                        argument_is_node = 1
            # Check if we get a node of type Node class
            if argument_is_node is not None:
                # Yes, we've got a node class so we can use it's configure function
                assert len(requests[k]) == 1
                current_request = requests[k][0]
                new_request = dsk_dict[k][argument_is_node].__self__.configure(
                    current_request
                )  # Call the class configuration function
                # if not isinstance(new_request, list):
                #     new_request = [new_request]
                # # convert back to dicts (here we are allowed to modify it)
                # new_request = [dict(r) for r in new_request]
            else:
                # We didn't get a Node class so there is no
                # custom configuration function: pass through
                new_request = {}
                assert len(requests[k]) == 1
                r = dict(requests[k][0])
                if r:
                    # sanitize request
                    r.pop("requires_request", None)
                    r.pop("insert_predecessor", None)
                    r.pop("clone_dependencies", None)
                    r.pop("remove_dependency", None)
                    r.pop("remove_dependenies", None)
                    new_request = r

            # update dependencies
            # we're going to get all dependencies of this node
            # then, we'll check if this node requires to clone it's input path
            # If so, each cloned path gets a different request from this node
            # (these are contained in a list in `clone_dependencies`)
            # we are going to introduce new keys and new nodes in the graph
            # therefore we must update this nodes input keys (hacking it from/to dsk_dict[k][DATA])
            # for each clone

            # For now it's not possible to have predecessors and multiple requests
            # User must use one request with `clone_dependencies` and `insert_predecessor` keys
            insert_predecessor = []
            if (
                new_request is not None
                and not isinstance(new_request, list)
                and "insert_predecessor" in new_request
                and new_request["insert_predecessor"]
            ):
                insert_predecessor = new_request["insert_predecessor"]
                del new_request["insert_predecessor"]

            current_deps = get_dependencies(dsk_dict, k, as_list=True)
            k_in_keys = None
            if len(dsk_dict[k]) > DATA:
                k_in_keys = deepcopy(dsk_dict[k][DATA])  # [DATA] equals in_keys in dict

            if not isinstance(new_request, list):
                clone_dependencies = [new_request]
            else:
                clone_dependencies = new_request

            # check if any of our current dependencies already has to fulfil a request
            # since the request's might collide we should just duplicate it
            # in this run it gets a new name, and the existing one is left untouched until it's its turn.
            clone = False
            if clone_instead_merge:
                if len(clone_dependencies) > 1:
                    clone = True
                    k_in_keys = []
                else:
                    for d in current_deps:
                        if len(requests.get(d, [])) > 0:
                            clone = True
                            k_in_keys = []

            # if it's a list it automatically clones it, if its not
            # the user could use the clone_dependencies to clone it
            if (
                new_request is not None
                and not isinstance(new_request, list)
                and "clone_dependencies" in new_request
                and new_request["clone_dependencies"]
            ):
                clone_dependencies = new_request["clone_dependencies"]
                del new_request["clone_dependencies"]
                clone = True
                k_in_keys = []

            if (
                new_request is not None
                and not isinstance(new_request, list)
                and "requires_request" in new_request
                and new_request["requires_request"]
            ):
                del new_request["requires_request"]
                # input_requests[k] = NestedFrozenDict(new_request[0])
                input_requests[k] = new_request["self"]

            # all_deps = get_all_dependencies()
            clone_dependencies = [c for c in clone_dependencies if c is not None]

            for clone_id, request in enumerate(clone_dependencies):
                if clone:
                    to_clone_keys = dsk_dict[k][DATA]
                    if not isinstance(to_clone_keys, list):
                        to_clone_keys = [to_clone_keys]

                    # create new node in graph containing k_in_keys as input
                    if insert_predecessor:
                        pre_function = insert_predecessor[clone_id]
                        pre_request = clone_dependencies[clone_id]

                        pre_k = tokenize([k, "deklare_pre", clone_id])
                        pre_base_name = None
                        if hasattr(pre_function, "__self__") and hasattr(
                            pre_function.__self__, "dask_key_name"
                        ):
                            pre_base_name = pre_function.__self__.dask_key_name
                            pre_k = pre_base_name + KEY_SEP + pre_k

                        # # go trough request and update the name of the clone
                        # if pre_base_name is not None:
                        #     update_key_in_config(pre_request,pre_base_name,pre_k)

                        requests[pre_k] = [pre_request]
                        dsk_dict[pre_k] = [apply, pre_function, [], {}]
                        pre_in_keys = []

                    for to_clone_key in to_clone_keys:
                        if insert_predecessor:
                            if to_clone_key is None:
                                pre_in_keys.append(None)
                            else:
                                pre_in_keys.append(
                                    generate_clone_key(k, to_clone_key, clone_id)
                                )
                        else:
                            if to_clone_key is None:
                                k_in_keys.append(None)
                            else:
                                k_in_keys.append(
                                    generate_clone_key(k, to_clone_key, clone_id)
                                )

                    if insert_predecessor:
                        dsk_dict[pre_k][DATA] = pre_in_keys
                        dsk_dict[pre_k] = tuple(dsk_dict[pre_k])

                        k_in_keys += [pre_k]
                        remove[pre_k] = False
                        new_work[pre_k] = True
                        # new_work.append(pre_k)

                for i, d in enumerate(current_deps):
                    # duplicate keys

                    if clone:
                        clone_work = [d]

                        d = generate_clone_key(k, d, clone_id)
                        while clone_work:
                            new_clone_work = []
                            for cd in clone_work:
                                clone_d = generate_clone_key(k, cd, clone_id)

                                # update_key_in_config(request,cd,clone_d)
                                # TODO: do we need to reset the dask_key_name of each
                                #       of each cloned node?

                                normalize_node(cd)

                                cloned_cd_node = copy(dsk_dict[cd])

                                # if contains data as input
                                to_clone_keys = cloned_cd_node[DATA]
                                if not isinstance(to_clone_keys, list):
                                    to_clone_keys = [to_clone_keys]
                                cd_in_keys = []
                                for to_clone_key in to_clone_keys:
                                    if to_clone_key is None:
                                        cd_in_keys.append(None)
                                    else:
                                        cd_in_keys.append(
                                            generate_clone_key(
                                                k, to_clone_key, clone_id
                                            )
                                        )
                                # if len(cd_in_keys) == 1:
                                #     cd_in_keys = cd_in_keys[0]
                                nd = list(cloned_cd_node)
                                nd[DATA] = cd_in_keys
                                cloned_cd_node = tuple(nd)
                                dsk_dict[clone_d] = cloned_cd_node
                                new_deps = get_dependencies(dsk_dict, cd, as_list=True)
                                new_clone_work += new_deps
                            clone_work = new_clone_work

                    # determine what needs to be removed
                    if not insert_predecessor:
                        # we are not going to remove anything if we inserted a predecessor node before current node k
                        # we are also not updating the requests of dependencies of the original node k
                        # since it will be done in the next interaction by configuring the inserted predecessor

                        to_be_removed = False
                        if k in remove:
                            to_be_removed = remove[k]
                        if request is None:
                            to_be_removed = True
                        if "remove_dependencies" in request:
                            to_be_removed = request["remove_dependencies"]
                            del request["remove_dependencies"]

                        # TODO: so far this doesn't allow to clone dependencies and delete only one of them.
                        #       it might be irrelevant.
                        if request.get("remove_dependency", {}).get(
                            base_name(d), False
                        ):
                            to_be_removed = True
                            del request["remove_dependency"][base_name(d)]

                        if not request.get("remove_dependency", True):
                            # clean up if an empty dict still exists
                            del request["remove_dependency"]
                        if d in requests:
                            if clone_instead_merge:
                                raise InternalError(
                                    f"A duplicate request was found for {d} with the request {requests[d]}, set clone_instead_merge=False to allow this"
                                )
                            if not to_be_removed:
                                requests[d] += [request]
                            remove[d] = remove[d] and to_be_removed
                        else:
                            if not to_be_removed:
                                requests[d] = [request]
                            # if we received None
                            remove[d] = to_be_removed

                        # only configure each node once in a round!
                        # if d not in new_work and d not in work:
                        #     new_work.append(
                        #         d
                        #     )  # TODO: Do we need to configure dependency if we'll remove it?
                        # we should also add `d`` only to the work list if we did not insert a
                        # a predecessor. The predecessor will take care of adding it in the next round
                        # otherwise it could happen that the predecessor changes the name of the node
                        # by cloning it. Then we'd have a deprecated node name in the work list
                        if d not in work and (d not in remove or not remove[d]):
                            new_work[d] = True
            dsk_k = list(dsk_dict[k])
            # loop though all input keys of this node `k` and discard all inputs that have been removed
            if k_in_keys is not None:
                if not isinstance(k_in_keys, list):
                    k_in_keys = [k_in_keys]
                k_in_keys = [
                    k_in_keys_k
                    for k_in_keys_k in k_in_keys
                    if (not remove.get(k_in_keys_k, False))
                ]
                # if len(k_in_keys) == 1:
                #     k_in_keys = k_in_keys[0]
                dsk_k[DATA] = k_in_keys
            dsk_dict[k] = tuple(dsk_k)

        work = new_work

    # Assembling the configured new graph
    out = {k: dsk_dict[k] for k in out_keys if not remove[k]}

    def clean_request(x):
        """Removes any leftovers of deklare configuration"""
        for item in ["remove_dependencies", "clone_dependencies", "insert_predecessor"]:
            x.pop(item, None)

        return x

    # After we have acquired all requests we can input the required_requests as a input node to the requiring node
    # we assume that the last argument is the request
    for k in input_requests:
        if k not in out:
            continue
        # input_requests[k] = clean_request(input_requests[k])
        # Here we assume that we always receive the same tuple of (bound method, data, request)
        # If the interface changes this will break #TODO: check for all cases
        if isinstance(out[k][REQUEST], tuple):
            # FIXME: find a better inversion of unpack_collections().
            #        this is very fragile
            # Check if we've already got a request as argument
            # This is the case if our node will make use of a general config
            # Then the present request is updated with the configured one
            # We need to recreate the tuple/list elements though. (dask changed)
            # TODO: use a distinct request class
            if out[k][REQUEST][0] is dict:
                my_dict = {}
                # FIXME: it does not account for nested structures
                for item in out[k][REQUEST][1]:
                    if isinstance(item[1], tuple):
                        if item[1][0] is tuple:
                            item[1] = tuple(item[1][1])
                        elif item[1][0] is list:
                            item[1] = list(item[1][1])
                    my_dict[item[0]] = item[1]
                my_dict = {item[0]: item[1] for item in out[k][REQUEST][1]}
                my_dict.update(input_requests[k])
                out[k] = out[k][:REQUEST] + (my_dict,)
            else:
                # replace the last entry
                out[k] = out[k][:REQUEST] + (input_requests[k],)

        # # TODO: verify that we can ignore this case
        # elif isinstance(out[k][REQUEST], dict):
        #     out[k] = out[k][:REQUEST] + (copy(out[k][REQUEST]) | copy(input_requests[k]),)
        else:
            # replace the last entry
            out[k] = out[k][:REQUEST] + (input_requests[k],)

        # TODO: we might dask.delayed(out[k][REQUEST]) here

    # convert to delayed object
    in_keys = list(flatten(keys))

    def dask_optimize(dsk, keys):
        dsk1, deps = cull(dsk, keys)
        dsk2 = inline(dsk1, dependencies=deps)
        # dsk3 = inline_functions(dsk2, keys, [len, str.split],
        #                        dependencies=deps)
        dsk4, deps = fuse(dsk2)
        return dsk4

    #    dsk, dsk_keys = dask.base._extract_graph_and_keys([collection])

    # out = optimize_functions(out, in_keys)
    #    collection = Delayed(key=dsk_keys, dsk=collection)

    if len(in_keys) > 1:
        # collection = [Delayed(key=key, dsk=out) for key in in_keys]
        collection = Delayed(key=in_keys, dsk=out)
    else:
        collection = Delayed(key=in_keys[0], dsk=out)
        if isinstance(delayed, list):
            collection = [collection]

    # if isinstance(delayed, list):
    #     collection = [collection]

    if optimize_graph:
        collection = optimize(collection, keys)
    #
    return collection


def dfs_order(dsk, keys=None, dask_optimize=False):
    """performs a depth first search and orders the items accordingly"""

    # if not isinstance(delayed, list):
    #     collections = [delayed]
    # else:
    #     collections = delayed

    # dsk, dsk_keys = dask.base._extract_graph_and_keys(collections)

    # dsk_dict = {k: dsk[k] for k in dsk.keys()}
    dsk_dict = dsk

    dependencies = {k: get_dependencies(dsk, k) for k in dsk}
    dependents = reverse_dict(dependencies)
    root_nodes = {k for k, v in dependents.items() if not v}
    keys = root_nodes

    if not isinstance(keys, (list, set)):
        keys = [keys]

    work = list(set(flatten(keys)))

    def dfs(k, dsk_dict, keyorder, idx):
        current_deps = get_dependencies(dsk_dict, k, as_list=True)
        for i, d in enumerate(current_deps):
            if d not in keyorder:
                keyorder, idx = dfs(d, dsk_dict, keyorder, idx)
                keyorder[d] = idx
                idx = idx + 1

        return keyorder, idx

    keyorder = {}
    idx = 0
    for k in work:
        keyorder, idx = dfs(k, dsk_dict, keyorder, idx)
        if k not in keyorder:
            keyorder[k] = idx
            idx += 1

    return keyorder


def optimize(delayed, keys=None, dask_optimize=False):
    """Optimizes the graph after configuration"""

    # TODO: Why is this necessary?
    if not isinstance(delayed, list):
        collections = [delayed]
    else:
        collections = delayed

    dsk, dsk_keys = dask.base._extract_graph_and_keys(collections)

    dsk_dict = {k: dsk[k] for k in dsk.keys()}

    if keys is None:
        keys = dsk_keys

    if not isinstance(keys, (list, set)):
        keys = [keys]

    work = list(set(flatten(keys)))

    # Invert the task graph: make a compute graph
    dsk_inv = {k: {} for k in work}

    out_keys = []
    sources = set()
    seen = set()
    while work:
        new_work = []

        out_keys += work
        for k in work:
            current_deps = get_dependencies(dsk_dict, k, as_list=True)

            if not current_deps:
                sources.add(k)

            for i, d in enumerate(current_deps):
                if d not in dsk_inv:
                    dsk_inv[d] = {k}
                else:
                    dsk_inv[d].add(k)

                if d not in seen:
                    new_work.append(d)
                    seen.add(d)

        work = new_work

    def replace(s, r, n):
        if isinstance(s, str) and s == r:
            return n
        return s

    # traverse the task graph in compute direction

    # starting from the sources
    work = list(sources)
    in_keys = list(set(flatten(keys)))
    out_keys = []
    seen = set()
    # import time
    # debug={}
    rename = {}
    while work:
        new_work = []

        for k in work:
            if k in in_keys:
                # if we are at a sink node we don't change the name
                out_keys += [k]
                continue

            # rename k
            node_token = ""

            # FIXME: for which cases do we need the following two lines?
            # it breaks the optimization significantly for chunkpersister, because it initializes
            # a new hashpersister instance for each parallel branch dynamically
            # What is the purpose? Do we want to account for internal configurations? Maybe find a different solution
            # if hasattr(dsk_dict[k][0], "__self__"):
            #     node_token = dsk_dict[k][0].__self__

            input_token = list(dsk_dict[k][1:])

            token = tokenize([node_token, input_token])

            # start = time.time()
            new_k = base_name(k) + KEY_SEP + token
            # duration = time.time() - start
            # debug[base_name(k)] = debug.get(base_name(k),0) + duration

            out_keys += [new_k]
            rename[k] = new_k

            current_dependants = [d for d in dsk_inv[k]]

            for i, d in enumerate(current_dependants):
                # TODO: is there a way to not change the dsk_dict in-place?
                if isinstance(dsk_dict[d][DATA], list):
                    in1 = [replace(s, k, new_k) for s in dsk_dict[d][DATA]]
                elif isinstance(dsk_dict[d][DATA], str):
                    in1 = replace(dsk_dict[d][DATA], k, new_k)
                else:
                    in1 = dsk_dict[d][DATA]

                dsk_dict[d] = tuple(
                    list(dsk_dict[d][:DATA]) + [in1] + list(dsk_dict[d][3:])
                )

                if d not in seen:
                    new_work.append(d)
                    seen.add(d)

        work = new_work

    for k in rename:
        dsk_dict[rename[k]] = dsk_dict[k]

    out = {k: dsk_dict[k] for k in out_keys}

    def optimize_functions(dsk, keys):
        dsk1, deps = cull(dsk, keys)
        dsk2 = inline(dsk1, dependencies=deps)
        # dsk3 = inline_functions(dsk2, keys, [len, str.split],
        #                        dependencies=deps)
        dsk4, deps = fuse(dsk2, fuse_subgraphs=True)
        return dsk4

    if dask_optimize:
        out = optimize_functions(out, in_keys)

    if len(in_keys) > 1:
        # collection = [Delayed(key=key, dsk=out) for key in in_keys]
        collection = Delayed(key=in_keys, dsk=out)
    else:
        collection = Delayed(key=in_keys[0], dsk=out)
        if isinstance(delayed, list):
            collection = [collection]

    return collection


def compute(graph, request):
    configured_graph = configuration(graph, request)
    # dsk, dsk_keys = dask.base._extract_graph_and_keys([configured_graph])

    computed_result = dask.compute(configured_graph)[0]
    return computed_result


class create_taskgraph:
    def __init__(self):
        self.prev = False
        pass

    def __enter__(self):
        if setting_exists("use_delayed"):
            self.prev = get_setting("use_delayed")
        set_setting("use_delayed", True)

    def __exit__(self, type, value, traceback):
        set_setting("use_delayed", self.prev)



"""
Copyright 2024 Swiss Federal Institute of Technology (ETH Zurich), Matthias Meyer

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import functools
import inspect
import warnings
from copy import copy, deepcopy
from uuid import uuid4

import dask
import dask.delayed
from dask.delayed import Delayed

KEY_SEP = "+"


class FlowContext:
    __context: dict = {}
    __enabled: bool = False

    @classmethod
    def get(name):
        return FlowContext.__context[name]

    @staticmethod
    def exists(name):
        return name in FlowContext.__context

    @staticmethod
    def set(name, value):
        FlowContext.__context[name] = value

    @staticmethod
    def is_enabled():
        return FlowContext.__enabled

    @staticmethod
    def set_enabled(value: bool):
        FlowContext.__enabled = value

    @staticmethod
    def reset():
        FlowContext.__enabled = False
        FlowContext.__context = {}


class create_taskgraph:
    def __init__(self):
        self.prev = False

    def __enter__(self):
        global is_enabled
        FlowContext.reset()
        FlowContext.set_enabled(True)
        is_enabled = True

    def __exit__(self, type, value, traceback):
        global is_enabled
        is_enabled = False
        FlowContext.reset()


def dict_update(base, update, convert_nestedfrozen=False):
    if not isinstance(base, dict) or not isinstance(update, dict):
        raise TypeError(
            f"dict_update requires two dicts as input. But we received {type(base)} and {type(update)}"
        )

    for key in update:
        if isinstance(base.get(key), dict) and isinstance(update[key], dict):
            if convert_nestedfrozen:
                base[key] = dict(base[key])
            base[key] = dict_update(
                base[key], update[key], convert_nestedfrozen=convert_nestedfrozen
            )
        else:
            base[key] = update[key]

    return base


is_enabled = False


class Node(object):
    def __init__(self, **kwargs):
        self.config = locals().copy()
        # FIXME: This works but there are better solutions!
        while "kwargs" in self.config:
            if "kwargs" not in self.config["kwargs"]:
                self.config.update(self.config["kwargs"])
                break
            self.config.update(self.config["kwargs"])

        del self.config["kwargs"]
        del self.config["self"]

        self._name = None

    def merge_config(self, request):
        """Each request contains configuration which may apply to different
        node instances. This function collects all information that apply to _this_
        node (including it's preset configs) and adds a `self` keyword to the request.

        Args:
            request (dict): The request and configuration options.

        Returns:
            dict: A new request which specific to this node.
        """
        protected_keys = ["self", "config"]

        new_request = {}
        new_request.update(request)
        new_request["self"] = {}
        if hasattr(self, "config"):
            new_request["self"].update(deepcopy(self.config))
        for key in request:
            # we consider any key that is not a protected key a global key
            if key not in protected_keys:
                new_request["self"][key] = request[key]

        if "config" in request:
            PROTECTED_CONFIG_KEYS = ["global", "types", "keys"]

            # new_request['config'] = {}

            # We'll go through all parameters in the request's config
            # and add them to the self parameters

            # We assume that anything within 'config' is global
            for key in request["config"]:
                if key in PROTECTED_CONFIG_KEYS:
                    continue
                new_request["self"][key] = request["config"][key]

                # # add to new_request
                # new_request['config'][key] = request["config"][key]

            # add specific global config entries
            if "global" in request["config"]:
                for key in request["config"]["global"]:
                    new_request["self"][key] = request["config"]["global"][key]

                # # add to new_request
                # new_request['config']['global'] = request["config"]["global"]

            # add type specific configs (overwrites global config)
            if "types" in request["config"]:
                if type(self).__name__ in request["config"]["types"]:
                    dict_update(
                        new_request["self"],
                        request["config"]["types"][type(self).__name__],
                    )

                # # add to new_request
                # new_request['config']['types'] = request["config"]["types"]

            # add key specific configs (overwrites global and type config)
            if "keys" in request["config"]:
                if self.dask_key_name in request["config"]["keys"]:
                    dict_update(
                        new_request["self"],
                        request["config"]["keys"][self.dask_key_name],
                    )

                    # TODO: It should be save to remove these keys from the new_request!?
                    del new_request["config"]["keys"][self.dask_key_name]

                # TODO: should we prefer the following way of removing the config?
                # new_request['config']['keys'] = {k:v for k,v in request["config"]["keys"].items() if k != self.dask_key_name}

        if "config" in request:
            new_request["config"] = request["config"]
        return new_request

    def configure(self, request):
        """Before a task graph is executed each node is configured.
            The request is propagated from the end to the beginning
            of the DAG and each nodes "configure" routine is called.
            The request can be updated to reflect additional requirements,
            The return value gets passed to predecessors.

            Essentially the following question must be answered within the
            nodes configure function:
            What do I need to fulfil the request of my successor? Either the node
            can provide what is required or the request is passed through to
            predecessors in hope they can fulfil the request.

            Here, you must not configure the internal parameters of the
            Node otherwise it would not be thread-safe. You can however
            introduce a new key 'requires_request' in the request being
            returned. This request will then be passed as an argument
            to the __call__ function.

            Best practice is to configure the Node on initialization with
            runtime independent configurations and define all runtime
            dependant configurations here.

        Args:
            requests {List} -- List of requests (i.e. dictionaries).


        Returns:
            dict -- The (updated) request. If updated, modifications
                    must be made on a copy of the input. The return value
                    must be a dictionary.
                    If multiple requests are input to this function they
                    must be merged.
                    If nothing needs to be requested an empty dictionary
                    can be return. This removes all dependencies of this
                    node from the task graph.

        """
        merged_request = self.merge_config(request)

        # set default
        merged_request["requires_request"] = True
        return merged_request

    def __call__(self, *args, **kwargs):
        name = kwargs.get("name", None)
        context = kwargs.get("context", None)
        if name is not None and KEY_SEP in name:
            raise RuntimeError(f"Do not use a `{KEY_SEP}` character in your {name=}")
        if name is None:
            name = self._name

        new_kwargs = copy(kwargs)
        if kwargs.get("request", None) is not None:
            new_kwargs["request"] = self.merge_config(kwargs["request"])
        elif kwargs.get("deskriptor", None) is not None:
            new_kwargs["deskriptor"] = self.merge_config(kwargs["deskriptor"])
        else:
            new_kwargs["request"] = self.merge_config({})

        if context is None:
            context = FlowContext

        forward_func = self.compute

        if context.is_enabled():
            func = dask.delayed(forward_func)(*args, dask_key_name=name, **kwargs)
            # func = dask.delayed(forward_func)(data=None, request=None,
            #     dask_key_name=name
            # )
            self.dask_key_name = func.key
            return func
        else:
            return forward_func(*args, **kwargs)


def it(flow):
    if inspect.isclass(flow):
        flow = flow()

    with create_taskgraph():
        flow_graph = flow()

    return flow_graph


def task(name=None, context=None):
    if context is None:
        context = FlowContext

    def decorator_task(func_or_cls):
        if inspect.isclass(func_or_cls):
            cls = func_or_cls

            if cls.__name__ == "DeklareClass":
                # don't wrap it twice!
                return cls

            class DeklareClass(cls, Node):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self._name = name

            # Check if the original class defines configure
            if "configure" in cls.__dict__:
                original_inherit_method = cls.__dict__["configure"]

                def new_configure(self, request):
                    request = Node.configure(self, request)
                    return original_inherit_method(self, request)

                setattr(DeklareClass, "configure", new_configure)

            # Rename cls's __call__ method to compute
            if "__call__" in cls.__dict__:
                setattr(DeklareClass, "compute", cls.__dict__["__call__"])
                # Use Node's __call__ method as NewClass's __call__ method
                setattr(DeklareClass, "__call__", Node.__call__)

            return DeklareClass
        else:
            func = func_or_cls

            if isinstance(func, Delayed):
                # don't wrap it twice!
                return func

            @functools.wraps(func)
            def wrapper_decorator(*args, **kwargs):
                if context.is_enabled():
                    if name is None:
                        key_name = func.__name__
                    else:
                        key_name = name

                    ext = ""
                    while context.exists(key_name + ext):
                        ext = uuid4().hex[-6:]

                    key_name = key_name + ext
                    context.set(key_name, func)

                    if ext != "":
                        warnings.warn(
                            f"Duplicate name detected. Name changed to {key_name}"
                        )

                    # make a graph node
                    return dask.delayed(func)(*args, dask_key_name=key_name, **kwargs)
                else:
                    # compute function and return result
                    return func(*args, **kwargs)

        return wrapper_decorator

    return decorator_task
